from __future__ import annotations

from collections import deque
from hikariwave.audio.source import AudioSource
from hikariwave.event.types import WaveEventType
from hikariwave.internal.constants import Audio
from hikariwave.internal.result import Result, ResultReason
from typing import Any, Callable, Coroutine, TYPE_CHECKING

import asyncio
import logging
import random
import struct
import time

if TYPE_CHECKING:
    from hikariwave.connection import VoiceConnection

__all__ = ("AudioPlayer",)

logger: logging.Logger = logging.getLogger("hikariwave.player")

class AudioPlayer:
    """Responsible for all audio."""

    __slots__ = (
        "_connection", "_buffer", "_ended", "_skip", "_resumed",
        "_sequence", "_timestamp", "_nonce",
        "_queue", "_history", "_priority_source", "_current",
        "_player_task", "_lock", "_track_completed", "_volume",
    )

    def __init__(self, connection: VoiceConnection, max_history: int = 20) -> None:
        """
        Create a new audio player.
        
        Parameters
        ----------
        connection : VoiceConnection
            The active voice connection.
        max_history : int
            Maximum number of tracks to keep in history.
        """
        
        self._connection: VoiceConnection = connection
        self._buffer: asyncio.Queue[bytes | None] = asyncio.Queue()

        self._ended: asyncio.Event = asyncio.Event()
        self._skip: asyncio.Event = asyncio.Event()
        self._resumed: asyncio.Event = asyncio.Event()
        self._resumed.set()

        self._sequence: int = 0
        self._timestamp: int = 0
        self._nonce: int = 0

        self._queue: deque[AudioSource] = deque()
        self._history: deque[AudioSource] = deque(maxlen=max_history)
        self._priority_source: AudioSource = None
        self._current: AudioSource = None

        self._player_task: asyncio.Task = None
        self._lock: asyncio.Lock = asyncio.Lock()

        self._track_completed: bool = False
        self._volume: float | str | None = None

    def _generate_rtp(self) -> bytes:
        header: bytearray = bytearray(12)
        header[0] = 0x80
        header[1] = 0x78
        struct.pack_into(">H", header, 2, self._sequence)
        struct.pack_into(">I", header, 4, self._timestamp)
        struct.pack_into(">I", header, 8, self._connection._ssrc)

        return bytes(header)

    async def _play_internal(self, source: AudioSource) -> bool:
        self._buffer = asyncio.Queue()
        self._ended.clear()
        self._skip.clear()
        self._track_completed = False

        source._volume = source._volume or self._volume

        await self._connection._gateway.set_speaking(True)
        await self._connection._client._ffmpeg.submit(source, self._buffer)
        
        self._connection._client._event_factory.emit(
            WaveEventType.AUDIO_BEGIN,
            self._connection._channel_id,
            self._connection._guild_id,
            source,
        )
        
        while self._buffer.empty(): await asyncio.sleep(0.05)

        frame_duration: float = Audio.FRAME_LENGTH / 1000
        frame_count: int = 0
        start_time: float = time.perf_counter()

        while not self._ended.is_set() and not self._skip.is_set():
            if not self._resumed.is_set():
                await self._send_silence()
                await self._resumed.wait()

                frame_count = 0
                start_time = time.perf_counter()
                continue

            opus: bytes = await self._buffer.get()
            if not opus:
                self._track_completed = True
                break

            header: bytes = self._generate_rtp()
            encrypted: bytes = self._connection._mode(self._connection._secret, self._nonce, header, opus)
            await self._connection._server.send(encrypted)

            self._sequence = (self._sequence + 1) % Audio.BIT_16U
            self._timestamp = (self._timestamp + Audio.SAMPLES_PER_FRAME) % Audio.BIT_32U
            frame_count += 1

            target: float = start_time + (frame_count * frame_duration)
            sleep: float = target - time.perf_counter()

            if sleep > 0:
                await asyncio.sleep(sleep)
            elif sleep < -0.020:
                logger.debug(f"Frame {frame_count} is {-sleep:.3f}s behind schedule")
        
        if self._skip.is_set() and not self._ended.is_set():
            self._track_completed = False

        await self._send_silence()
        await self._connection._gateway.set_speaking(False)

        return self._track_completed

    async def _player_loop(self) -> None:
        while True:
            source: AudioSource = None

            async with self._lock:
                if self._priority_source:
                    source = self._priority_source
                    self._priority_source = None
                elif self._queue:
                    source = self._queue.popleft()
                else:
                    self._current = None
                    self._player_task = None
                    return
            
                self._current = source
            
            completed: bool = await self._play_internal(source)

            async with self._lock:
                self._connection._client._event_factory.emit(
                    WaveEventType.AUDIO_END,
                    self._connection._channel_id,
                    self._connection._guild_id,
                    self._current,
                )

                if completed or (self._skip.is_set() and not self._ended.is_set()):
                    self._history.append(source)

    async def _send_silence(self) -> None:
        send: Callable[[bytes], Coroutine[Any, Any, None]] = self._connection._server.send
        for _ in range(5):
            await send(b"\xF8\xFF\xFE")

    async def add_queue(self, source: AudioSource) -> Result:
        """
        Add an audio source to the queue.
        
        Parameters
        ----------
        source : AudioSource
            The source of the audio to add.

        Returns
        -------
        Result
            If the operation was successful, with reason provided if otherwise.

        Raises
        ------
        TypeError
            If the provided source doesn't inherit `AudioSource`.
        """

        if not isinstance(source, AudioSource):
            error: str = "Provided audio source doesn't inherit from `AudioSource`"
            raise TypeError(error)

        async with self._lock:
            self._queue.append(source)

            if not self._player_task or self._player_task.done():
                self._player_task = asyncio.create_task(self._player_loop())
        
        return Result.succeeded()

    async def clear_queue(self) -> Result:
        """
        Clear all audio from the queue.

        Returns
        -------
        Result
            If the operation was successful, with reason provided if otherwise.
        """

        async with self._lock:
            if len(self._queue) < 1:
                return Result.failed(ResultReason.EMPTY_QUEUE)
            
            self._queue.clear()
        
        return Result.succeeded()

    @property
    def connection(self) -> VoiceConnection:
        """The active connection that is responsible for this player."""
        return self._connection

    @property
    def current(self) -> AudioSource | None:
        """The currently playing audio, if present."""
        return self._current

    @property
    def history(self) -> list[AudioSource]:
        """Get all audio previously played."""

        return list(self._history)

    @property
    def is_playing(self) -> bool:
        """If the player has audio currently playing."""
        return self._current is not None and self._resumed.is_set()

    async def next(self) -> Result:
        """
        Play the next audio in queue.

        Returns
        -------
        Result
            If the operation was successful, with reason provided if otherwise.
        """

        async with self._lock:
            if self._current is None:
                return Result.failed(ResultReason.NO_TRACK)
    
            if len(self._queue) < 1:
                return Result.failed(ResultReason.EMPTY_QUEUE)

            self._skip.set()
            self._resumed.set()
        
        return Result.succeeded()

    async def pause(self) -> Result:
        """
        Pause the current audio.

        Returns
        -------
        Result
            If the operation was successful, with reason provided if otherwise.
        """

        if self._current is None:
            return Result.failed(ResultReason.NO_TRACK)

        if not self._resumed.is_set():
            return Result.failed(ResultReason.PAUSED)

        self._resumed.clear()

        await self._connection._gateway.set_speaking(False)
        
        return Result.succeeded()

    async def play(self, source: AudioSource) -> Result:
        """
        Play audio from a source.
        
        Parameters
        ----------
        source : AudioSource
            The source of the audio to play
        
        Returns
        -------
        Result
            If the operation was successful, with reason provided if otherwise.

        Raises
        ------
        TypeError
            If the provided source doesn't inherit `AudioSource`.
        """

        if not isinstance(source, AudioSource):
            error: str = "Provided source must inherit `AudioSource`"
            raise TypeError(error)

        async with self._lock:
            self._priority_source = source

            if self._current is not None:
                self._skip.set()

            if not self._player_task or self._player_task.done():
                self._player_task = asyncio.create_task(self._player_loop())
        
        return Result.succeeded()

    async def previous(self) -> Result:
        """
        Play the latest previously played audio.

        Returns
        -------
        Result
            If the operation was successful, with reason provided if otherwise.
        """

        async with self._lock:
            if len(self._history) < 1:
                return Result.failed(ResultReason.EMPTY_HISTORY)
            
            previous: AudioSource = self._history.pop()
            self._queue.appendleft(previous)

            if self._current:
                self._skip.set()
                self._resumed.set()
            
        return Result.succeeded()

    @property
    def queue(self) -> list[AudioSource]:
        """Get all audio currently in queue."""

        return list(self._queue)

    async def remove_queue(self, source: AudioSource) -> Result:
        """
        Remove an audio source from the queue.
        
        Parameters
        ----------
        source : AudioSource
            The source of the audio to remove.
        
        Returns
        -------
        Result
            If the operation was successful, with reason provided if otherwise.
        
        Raises
        ------
        TypeError
            If the provided source doesn't inherit `AudioSource`.
        """

        if not isinstance(source, AudioSource):
            error: str = "Provided source must inherit `AudioSource`"
            raise TypeError(error)

        async with self._lock:
            try:
                self._queue.remove(source)
            except ValueError:
                return Result.failed(ResultReason.NOT_FOUND)
        
        return Result.succeeded()

    async def resume(self) -> Result:
        """
        Resume the current audio.

        Returns
        -------
        Result
            If the operation was successful, with reason provided if otherwise.
        """
        
        if self._resumed.is_set():
            return Result.failed(ResultReason.PLAYING)

        await self._connection._gateway.set_speaking(True)
        
        self._resumed.set()
        return Result.succeeded()

    def set_volume(self, volume: float | str | None = None) -> None:
        """
        Set the default volume of this player.
        Can be `None`, any scaled value (`1.0`, `2.0`, `0.5`, etc.) or dB-based (`-3dB`, `0.5dB`, etc.).
        
        Parameters
        ----------
        volume : float | str | None
            The volume to set as a default for this player.
        
        Raises
        ------
        TypeError
            If `volume` is provided and it's not `float`, `int`, or `str`.
        """

        if volume is not None and not isinstance(volume, (float, int, str)):
            error: str = "Provided volume must be a `float`, `int`, or `str`"
            raise TypeError(error)
        
        self._volume = volume

    async def shuffle(self) -> Result:
        """
        Shuffle all audio currently in queue.

        Returns
        -------
        Result
            If the operation was successful, with reason provided if otherwise.
        """

        async with self._lock:
            if len(self._queue) < 1:
                return Result.failed(ResultReason.EMPTY_QUEUE)
            
            temp: list[AudioSource] = list(self._queue)
            random.shuffle(temp)
            self._queue.clear()
            self._queue.extend(temp)
        
        return Result.succeeded()

    async def stop(self) -> Result:
        """
        Stop the current audio.

        Returns
        -------
        Result
            If the operation was successful, with reason provided if otherwise.
        """
        
        self._ended.set()
        self._skip.set()
        self._resumed.set()

        async with self._lock:
            self._queue.clear()
            self._priority_source = None
            self._current = None
        
        await self._connection._gateway.set_speaking(False)
        
        return Result.succeeded()
    
    @property
    def volume(self) -> float | str | None:
        """If set, the player's default volume."""
        return self._volume


from __future__ import annotations

from enum import auto, IntEnum
from hikariwave.audio.player import AudioPlayer
from hikariwave.event.types import WaveEventType
from hikariwave.internal.encrypt import Encrypt
from hikariwave.networking.gateway import Opcode, ReadyPayload, SessionDescriptionPayload, VoiceGateway
from hikariwave.networking.server import VoiceServer
from typing import Callable, TYPE_CHECKING

import asyncio
import hikari

if TYPE_CHECKING:
    from hikariwave.client import VoiceClient

__all__ = ("VoiceConnection",)

class ConnectionStatus(IntEnum):
    """Represents the current state of a voice connection."""

    CONNECTED = auto()
    """Connected to voice gateway and voice server."""
    CONNECTING = auto()
    """Connecting to voice gateway and voice server."""
    DISCONNECTED = auto()
    """Disconnected from voice gateway and voice server."""
    NEW = auto()
    """Instantiated and waiting to begin connection."""
    RECONNECTING = auto()
    """Reconnecting to voice gateway and voice server."""

class VoiceConnection:
    """An active connection to a Discord voice channel."""

    __slots__ = (
        "_client", "_guild_id", "_channel_id", "_endpoint", "_session_id", "_token",
        "_server", "_gateway", "_ready", "_state", "_ssrc", "_mode", "_secret", "_player",
    )

    def __init__(
        self,
        client: VoiceClient,
        guild_id: hikari.Snowflakeish,
        channel_id: hikari.Snowflakeish,
        endpoint: str,
        session_id: str,
        token: str,
    ) -> None:
        """
        Create a new voice connection.
        
        Parameters
        ----------
        client : VoiceClient
            The controlling client for all connections and state.
        guild_id : hikari.Snowflakeish
            The ID of the guild the channel is in.
        channel_id : hikari.Snowflakeish
            The ID of the channel to connect to.
        endpoint : str
            The URL of Discord's voice gateway.
        session_id : str
            The provided session ID from Discord's OAuth2 gateway.
        token : str
            The provided token from Discord's OAuth2 gateway.
        """
        
        self._client: VoiceClient = client
        self._guild_id: hikari.Snowflakeish = guild_id
        self._channel_id: hikari.Snowflakeish = channel_id
        self._endpoint: str = endpoint
        self._session_id: str = session_id
        self._token: str = token

        self._server: VoiceServer = VoiceServer(self._client)
        self._gateway: VoiceGateway = VoiceGateway(
            self,
            self._guild_id,
            self._channel_id,
            self._client.bot.get_me().id,
            self._session_id,
            self._token,
        )
        self._gateway.set_callback(Opcode.READY, self._gateway_ready)
        self._gateway.set_callback(Opcode.SESSION_DESCRIPTION, self._gateway_session_description)
        self._ready: asyncio.Event = asyncio.Event()
        self._state: ConnectionStatus = ConnectionStatus.NEW

        self._ssrc: int = None
        self._mode: Callable[[bytes, int, bytes, bytes], bytes] = None
        self._secret: bytes = None

        self._player: AudioPlayer = AudioPlayer(self)
    
    async def _connect(self) -> None:
        if self._state in (ConnectionStatus.CONNECTED, ConnectionStatus.CONNECTING):
            return
        
        self._state = ConnectionStatus.CONNECTING
        self._ready.clear()

        try:
            await self._gateway.connect(f"{self._endpoint}/?v=8")
            await self._ready.wait()

            if self._state == ConnectionStatus.CONNECTING:
                self._state = ConnectionStatus.CONNECTED
        except Exception:
            self._state = ConnectionStatus.NEW
            raise

    async def _disconnect(self) -> None:
        if self._state == ConnectionStatus.DISCONNECTED:
            return
        
        self._state = ConnectionStatus.DISCONNECTED

        if self._player:
            await self._player.stop()
        
        if self._server:
            await self._server.disconnect()
    
        if self._gateway:
            await self._gateway.disconnect()

    async def _gateway_ready(self, payload: ReadyPayload) -> None:
        self._ssrc = payload.ssrc
        
        supported_modes: list[str] = payload.modes
        chosen_mode: str = None
        for mode in supported_modes:
            if mode not in Encrypt.SUPPORTED:
                continue

            if not hasattr(Encrypt, mode):
                continue

            chosen_mode = mode
            break

        if not chosen_mode:
            error: str = "No supported encryption method was found/implemented"
            raise RuntimeError(error)

        ip, port = await self._server.connect(payload.ip, payload.port, self._ssrc)

        await self._gateway.select_protocol(ip, port, chosen_mode)

    async def _gateway_reconnect(self) -> None:
        if self._state == ConnectionStatus.RECONNECTING:
            return
        
        self._state = ConnectionStatus.RECONNECTING
        self._ready.clear()

        if self._player and self._player.is_playing:
            await self._player.pause()
    
        await self._server.disconnect()
        await self._gateway.disconnect()

        self._server = VoiceServer(self._client)
        self._gateway = VoiceGateway(
            self,
            self._guild_id,
            self._channel_id,
            self._client.bot.get_me().id,
            self._session_id,
            self._token,
        )
        self._gateway.set_callback(Opcode.READY, self._gateway_ready)
        self._gateway.set_callback(Opcode.SESSION_DESCRIPTION, self._gateway_session_description)
        await self._gateway.connect(f"{self._endpoint}/?v=8")

        self._client._event_factory.emit(
            WaveEventType.VOICE_RECONNECT,
            self._channel_id,
            self._guild_id,
        )

    async def _gateway_session_description(self, payload: SessionDescriptionPayload) -> None:
        self._mode = getattr(Encrypt, payload.mode)
        self._secret = payload.secret
        self._state = ConnectionStatus.CONNECTED

        self._ready.set()

        if not self._player._resumed.is_set() and self._player._current:
            await self._player.resume()

    @property
    def channel_id(self) -> hikari.Snowflakeish:
        """The ID of the channel this connection is in."""
        return self._channel_id

    @property
    def client(self) -> hikari.GatewayBot:
        """The controlling OAuth2 bot."""
        return self._client

    async def disconnect(self) -> None:
        """
        Disconnect from the current channel.
        """
        
        await self._client.disconnect(self._guild_id)
    
    @property
    def guild_id(self) -> hikari.Snowflakeish:
        """The ID of the guild this connection is in."""
        return self._guild_id

    @property
    def latency_gateway(self) -> float | None:
        """Get the heartbeat latency of this connection with Discord's gateway, if connected."""
        
        if not self._gateway._last_heartbeat_ack:
            return None
        
        return self._gateway._last_heartbeat_ack - self._gateway._last_heartbeat_sent
    
    @property
    def player(self) -> AudioPlayer:
        """The audio player associated with this connection."""
        return self._player



from __future__ import annotations

from cryptography.hazmat.primitives.ciphers.aead import AESGCM
from hikariwave.internal.constants import Audio

import nacl.secret as secret
import struct

__all__ = ("Encrypt",)

class Encrypt:
    """Container class for all supported, non-deprecated encryption modes."""

    SUPPORTED: tuple[str, ...] = ("aead_aes256_gcm_rtpsize", "aead_xchacha20_poly1305_rtpsize",)
    """A list of all currently supported, non-deprecated, complete, and tested encryption modes."""

    @staticmethod
    def aead_aes256_gcm_rtpsize(secret_key: bytes, nonce: int, header: bytes, audio: bytes) -> bytes:
        """
        Encrypt audio using `aead_aes256_gcm_rtpsize`.
        
        Parameters
        ----------
        secret_key : bytes
            32-byte AES encryption key provided by Discord.
        nonce : int
            32-bit packet counter.
        header : bytes
            RTP header (12 bytes).
        audio : bytes
            Opus audio payload.
        
        Returns
        -------
        bytes
            The encrypted audio packet.
        """

        aesgcm: AESGCM = AESGCM(secret_key)

        packet_nonce: bytes = struct.pack(">I", nonce) + b"\x00" * 8
        nonce = (nonce + 1) % Audio.BIT_32U
        
        encrypted: bytes = aesgcm.encrypt(packet_nonce, audio, header)

        return header + encrypted + packet_nonce[8:]

    @staticmethod
    def aead_xchacha20_poly1305_rtpsize(secret_key: bytes, nonce: int, header: bytes, audio: bytes) -> bytes:
        """
        Encrypt audio using `aead_xchacha20_poly1305_rtpsize`.
        
        Parameters
        ----------
        secret_key : bytes
            32-byte AES encryption key provided by Discord.
        nonce : int
            32-bit packet counter.
        header : bytes
            RTP header (12 bytes).
        audio : bytes
            Opus audio payload.
        
        Returns
        -------
        bytes
            The encrypted audio packet.
        """
        
        box: secret.Aead = secret.Aead(secret_key)

        packet_nonce: bytearray = bytearray(24)
        packet_nonce[:4] = struct.pack(">I", nonce)
        nonce = (nonce + 1) % Audio.BIT_32U

        return header + box.encrypt(audio, header, bytes(packet_nonce)).ciphertext + packet_nonce[:4]


from __future__ import annotations

__all__ = (
    "ClientError",
    "GatewayError",
    "ServerError",
)

class ClientError(Exception):
    """Raised when an error occurs with a voice client."""

class GatewayError(Exception):
    """Raised when an error occurs with a voice system gateway."""

class ServerError(Exception):
    """Raised when an error occurs with a voice system server."""


from __future__ import annotations

from enum import auto, IntEnum

__all__ = (
    "VoiceWarningType",
    "WaveEventType",
)

class VoiceWarningType(IntEnum):
    """A type of voice warning."""

    LATENCY          = auto()
    """Voice server connection has a high latency."""
    UNHANDLED_OPCODE = auto()
    """Voice gateway received an unhandled operation code."""

class WaveEventType(IntEnum):
    """A type of supplemental event."""

    ALL                   = auto()
    """Base event listener - Receives all events."""
    AUDIO_BEGIN           = auto()
    """When audio begins playing."""
    AUDIO_END             = auto()
    """When audio stops playing."""
    BOT_JOIN_VOICE        = auto()
    """When the bot joins a channel."""
    BOT_LEAVE_VOICE       = auto()
    """When the bot leaves a channel."""
    MEMBER_DEAF           = auto()
    """When a member deafens/undeafens."""
    MEMBER_JOIN_VOICE     = auto()
    """When a member joins a channel."""
    MEMBER_LEAVE_VOICE    = auto()
    """When a member leaves a channel."""
    MEMBER_MOVE_VOICE     = auto()
    """When a member moves channels."""
    MEMBER_MUTE           = auto()
    """When a member mutes/unmutes."""
    MEMBER_START_SPEAKING = auto()
    """When a member starts speaking."""
    MEMBER_STOP_SPEAKING  = auto()
    """When a member stops speaking."""
    VOICE_RECONNECT       = auto()
    """When a voice connection has to reconnect."""
    VOICE_WARNING         = auto()
    """When a voice connection issues a warning."""


from __future__ import annotations

from dataclasses import dataclass
from hikariwave.audio.ffmpeg import FFmpegPool
from hikariwave.connection import VoiceConnection
from hikariwave.event.factory import EventFactory
from hikariwave.event.types import WaveEventType
from hikariwave.internal.constants import Audio
from hikariwave.internal.error import GatewayError
from typing import TypeAlias

import asyncio
import hikari
import logging

__all__ = ("VoiceClient",)

logger: logging.Logger = logging.getLogger("hikari-wave.client")

ChannelID: TypeAlias = hikari.Snowflakeish
GuildID: TypeAlias = hikari.Snowflakeish
MemberID: TypeAlias = hikari.Snowflakeish

@dataclass(slots=True)
class VoiceChannelMeta:
    """Metadata container for a voice channel."""

    active: set[hikari.Snowflake]
    """A set of all actively speaking users in this channel."""
    guild_id: hikari.Snowflake
    """The ID of the guild this channel is in."""
    id: hikari.Snowflake
    """The ID of this channel."""
    members: dict[hikari.Snowflake, hikari.Member]
    """All members inside of this channel."""

class VoiceClient:
    """Voice system implementation for `hikari`-based Discord bots."""

    __slots__ = (
        "_bot", "_audio_bitrate", "_audio_channels", "_audio_frame_length",
        "_connections", "_connectionsr", "_channels", "_members",
        "_ssrcs", "_ssrcsr", "_states", "_event_factory", "_ffmpeg",
    )

    def __init__(
        self,
        bot: hikari.GatewayBot,
        *,
        bitrate: str = Audio.BITRATE,
        channels: int = Audio.CHANNELS,
    ) -> None:
        """
        Create a new voice client.
        
        Parameters
        ----------
        bot : hikari.GatewayBot
            The `hikari`-based Discord bot to link this voice system with.
        bitrate : str
            The audio output bitrate.
        channels : int
            The amount of audio channels.
        
        Raises
        ------
        TypeError
            If `bot` isn't a `hikari.GatewayBot`.
        """

        if not isinstance(bot, hikari.GatewayBot):
            error: str = "Provided bot must be a `hikari.GatewayBot`"
            raise TypeError(error)
        
        self._bot: hikari.GatewayBot = bot
        self._bot.subscribe(hikari.VoiceStateUpdateEvent, self._disconnected)
        self._bot.subscribe(hikari.VoiceStateUpdateEvent, self._voice_state_update)

        self._audio_bitrate: str = bitrate
        self._audio_channels: int = channels

        self._connections: dict[GuildID, VoiceConnection] = {}
        self._connectionsr: dict[ChannelID, GuildID] = {}

        self._channels: dict[ChannelID, VoiceChannelMeta] = {}
        self._members: dict[MemberID, ChannelID] = {}
        self._ssrcs: dict[MemberID, int] = {}
        self._ssrcsr: dict[int, MemberID] = {}

        self._states: dict[MemberID, tuple[bool, bool]] = {}

        self._event_factory: EventFactory = EventFactory(self._bot)
        self._ffmpeg: FFmpegPool = FFmpegPool(self)
    
    async def _connect(self, guild_id: hikari.Snowflake, channel_id: hikari.Snowflake, mute: bool, deaf: bool, disconnect: bool = False) -> VoiceConnection:
        try:
            if disconnect:
                await self._disconnect(guild_id)
            
            if guild_id in self._connections and not disconnect:
                return self._connections[guild_id]

            logger.info(f"Connecting to voice: Guild={guild_id}, Channel={channel_id}, Mute={mute}, Deaf={deaf}")

            await self._bot.update_voice_state(guild_id, channel_id, self_mute=mute, self_deaf=deaf)

            try:
                server_update, state_update = await asyncio.gather(
                    self._bot.wait_for(
                        hikari.VoiceServerUpdateEvent, 3.0,
                        lambda e: e.guild_id == guild_id
                    ),
                    self._bot.wait_for(
                        hikari.VoiceStateUpdateEvent, 3.0,
                        lambda e: e.guild_id == guild_id and e.state.channel_id == channel_id and e.state.user_id == self._bot.get_me().id
                    )
                )
            except asyncio.TimeoutError:
                error: str = "Voice server/state update timed out"
                raise GatewayError(error)

            guild: hikari.Guild = await self._bot.rest.fetch_guild(guild_id)
            members: dict[hikari.Snowflake, hikari.Member] = {}

            for state in guild.get_voice_states().values():
                if state.user_id == self._bot.get_me().id or state.channel_id != channel_id:
                    continue

                self._states[state.member.id] = (
                    state.is_guild_deafened or state.is_self_deafened,
                    state.is_guild_muted or state.is_self_muted,
                )

                members[state.member.id] = state.member
                self._members[state.member.id] = channel_id
            
            self._channels[channel_id] = VoiceChannelMeta(set(), guild_id, channel_id, members)

            connection: VoiceConnection = VoiceConnection(
                self,
                guild_id,
                channel_id,
                server_update.endpoint,
                state_update.state.session_id,
                server_update.token,
            )
            await connection._connect()
            self._connections[guild_id] = connection
            self._connectionsr[channel_id] = guild_id

            self._event_factory.emit(
                WaveEventType.BOT_JOIN_VOICE,
                self._bot,
                channel_id,
                guild_id,
                deaf,
                mute,
            )

            return connection
        except Exception:
            await self._disconnect(guild_id)
            raise

    async def _disconnect(self, guild_id: hikari.Snowflake) -> None:
        if guild_id not in self._connections:
            return
        
        connection: VoiceConnection = self._connections.pop(guild_id)

        logger.info(f"Disconnecting from voice: Guild={guild_id}, Channel={connection._channel_id}")

        del self._connectionsr[connection._channel_id]
        await connection._disconnect()

        cmeta: VoiceChannelMeta = self._channels.pop(connection._channel_id, None)
        if cmeta:
            for member_id in cmeta.members.keys():
                del self._members[member_id]
                ssrc: int | None = self._ssrcs.pop(member_id, None)
                if ssrc:
                    self._ssrcsr.pop(ssrc, None)

        self._event_factory.emit(
            WaveEventType.BOT_LEAVE_VOICE,
            self._bot,
            connection._channel_id,
            guild_id,
        )

    async def _disconnected(self, event: hikari.VoiceStateUpdateEvent) -> None:
        if event.state.user_id != self._bot.get_me().id:
            return
        
        if event.guild_id not in self._connections:
            return
        
        await self._disconnect(event.guild_id)

    async def _voice_state_update(self, event: hikari.VoiceStateUpdateEvent) -> None:
        state: hikari.VoiceState = event.state
        member: hikari.Member = state.member
        if state.user_id == self._bot.get_me().id or not member:
            return
        
        guild_id: hikari.Snowflake = state.guild_id
        old_channel_id: hikari.Snowflake | None = self._members.get(member.id)
        new_channel_id: hikari.Snowflake | None = state.channel_id

        # Member Joined Channel
        if new_channel_id and not old_channel_id:
            member.is_deaf = state.is_guild_deafened or state.is_self_deafened
            member.is_mute = state.is_guild_muted or state.is_self_muted
            self._states[member.id] = (member.is_deaf, member.is_mute)

            if new_channel_id in self._channels:
                cmeta: VoiceChannelMeta = self._channels[new_channel_id]
                cmeta.members[member.id] = member
                self._members[member.id] = new_channel_id

            self._event_factory.emit(
                WaveEventType.MEMBER_JOIN_VOICE,
                new_channel_id,
                guild_id,
                member,
            )
        # Member Moved Channels
        elif new_channel_id and old_channel_id and old_channel_id != new_channel_id:
            if old_channel_id in self._channels:
                cmeta: VoiceChannelMeta = self._channels[old_channel_id]
                del cmeta.members[member.id]
                del self._members[member.id]

                ssrc: int | None = self._ssrcs.pop(member.id, None)
                if ssrc:
                    del self._ssrcsr[ssrc]

            if new_channel_id in self._channels:
                cmeta: VoiceChannelMeta = self._channels[new_channel_id]
                cmeta.members[member.id] = member
                self._members[member.id] = new_channel_id

                ssrc: int | None = self._ssrcs.pop(member.id, None)
                if ssrc:
                    del self._ssrcsr[ssrc]

            self._event_factory.emit(
                WaveEventType.MEMBER_MOVE_VOICE,
                guild_id,
                member,
                new_channel_id,
                old_channel_id,
            )
        # Member Left Channel
        elif not new_channel_id and old_channel_id:
            self._states.pop(member.id, None)

            if old_channel_id in self._channels:
                cmeta: VoiceChannelMeta = self._channels[old_channel_id]
                del cmeta.members[member.id]
                del self._members[member.id]

                ssrc: int | None = self._ssrcs.pop(member.id, None)
                if ssrc:
                    del self._ssrcsr[ssrc]

            self._event_factory.emit(
                WaveEventType.MEMBER_LEAVE_VOICE,
                old_channel_id,
                guild_id,
                member,
            )
        # Member Update
        elif new_channel_id and old_channel_id and new_channel_id == old_channel_id:
            member.is_deaf = state.is_guild_deafened or state.is_self_deafened
            member.is_mute = state.is_guild_muted or state.is_self_muted

            old_deaf, old_mute = self._states[member.id]

            if new_channel_id in self._channels:
                cmeta: VoiceChannelMeta = self._channels[new_channel_id]
                cmeta.members[member.id] = member

            if old_deaf != member.is_deaf:
                self._event_factory.emit(
                    WaveEventType.MEMBER_DEAF,
                    new_channel_id,
                    guild_id,
                    member,
                    member.is_deaf,
                )
            
            if old_mute != member.is_mute:
                self._event_factory.emit(
                    WaveEventType.MEMBER_MUTE,
                    new_channel_id,
                    guild_id,
                    member,
                    member.is_mute,
                )
            
            self._states[member.id] = (member.is_deaf, member.is_mute)

    @property
    def bot(self) -> hikari.GatewayBot:
        """The controlling OAuth2 bot."""
        return self._bot

    async def close(self) -> None:
        """
        Shut down every connection and clean up.
        """

        self._bot.unsubscribe(hikari.VoiceStateUpdateEvent, self._disconnected)
        self._bot.unsubscribe(hikari.VoiceStateUpdateEvent, self._voice_state_update)

        await asyncio.gather(
            *(self._disconnect(guild_id) for guild_id in self._connections.keys())
        )

        if self._connections: self._connections.clear()
        if self._connectionsr: self._connectionsr.clear()

        if self._channels: self._channels.clear()
        if self._members: self._members.clear()
        if self._ssrcs: self._ssrcs.clear()
        if self._ssrcsr: self._ssrcsr.clear()

        if self._states: self._states.clear()

        await self._ffmpeg.stop()

    async def connect(
        self,
        guild_id: hikari.Snowflakeish,
        channel_id: hikari.Snowflakeish,
        *,
        mute: bool = False,
        deaf: bool = True,
    ) -> VoiceConnection:
        """
        Connect to a voice channel.
        
        Parameters
        ----------
        guild_id : hikari.Snowflakeish
            The ID of the guild that the channel is in.
        channel_id : hikari.Snowflakeish
            The ID of the channel to connect to.
        mute : bool 
            If the bot should be muted upon joining the channel.
        deaf : bool
            If the bot should be deafened upon joining the channel.
        
        Returns
        -------
        VoiceConnection
            The active connection to the voice channel, once fully connected.
        
        Raises
        ------
        asyncio.TimeoutError
            If Discord doesn't send a corresponding voice server/state update.
        TypeError
            - If `guild_id` or `channel_id` aren't `hikari.Snowflakeish`.
            - If `mute` or `deaf` aren't `bool`.
        """

        if not isinstance(guild_id, hikari.Snowflakeish):
            error: str = "Provided guild ID must be of type `hikari.Snowflakeish`"
            raise TypeError(error)
        
        if not isinstance(channel_id, hikari.Snowflakeish):
            error: str = "Provided channel ID must be of type `hikari.Snowflakeish`"
            raise TypeError(error)
        
        if not isinstance(mute, bool):
            error: str = "Provided mute state must be of type `bool`"
            raise TypeError(error)
        
        if not isinstance(deaf, bool):
            error: str = "Provided deaf state must be of type `bool`"
            raise TypeError(error)

        return await self._connect(guild_id, channel_id, mute, deaf)
    
    @property
    def connections(self) -> dict[hikari.Snowflake, VoiceConnection]:
        """A mapping of all voice connections."""
        return dict(self._connections)

    async def disconnect(
        self,
        *,
        guild_id: hikari.Snowflakeish | None = None,
        channel_id: hikari.Snowflakeish | None = None,
    ) -> None:
        """
        Disconnect from a voice channel.
        
        Parameters
        ----------
        guild_id : hikari.Snowflakeish | None
            The ID of the guild that the channel to disconnect from is in.
        channel_id : hikari.Snowflakeish | None
            The ID of the channel to disconnect from.
        
        Note
        ----
        At least one of `guild_id` or `channel_id` must be provided.

        Raises
        ------
        TypeError
            If `guild_id` or `channel_id` aren't `hikari.Snowflakeish`.
        ValueError
            If neither of `guild_id` or `channel_id` are provided.
        """
        
        if not guild_id and not channel_id:
            error: str = "At least guild_id or channel_id must be defined"
            raise ValueError(error)
        
        if guild_id and not isinstance(guild_id, hikari.Snowflakeish):
            error: str = "Provided guild ID must be of type `hikari.Snowflakeish`"
            raise TypeError(error)
        
        if channel_id and not isinstance(channel_id, hikari.Snowflakeish):
            error: str = "Provided channel ID must be of type `hikari.Snowflakeish`"
            raise TypeError(error)
        
        if channel_id:
            guild_id = self._connectionsr[channel_id]

        await self._bot.update_voice_state(guild_id, None)
        await self._disconnect(guild_id)
    
    def get_connection(
        self,
        *,
        guild_id: hikari.Snowflakeish | None = None,
        channel_id: hikari.Snowflakeish | None = None,
    ) -> VoiceConnection | None:
        """
        Get an active voice connection.
        
        Parameters
        ----------
        guild_id : hikari.Snowflakeish | None
            The ID of the guild that the connection is handling.
        channel_id : hikari.Snowflakeish | None
            The ID of the channel that the connection is handling.
        
        Note
        ----
        At least one of `guild_id` or `channel_id` must be provided.
        
        Returns
        -------
        VoiceConnection | None
            The active voice connection at the guild/channel, if present.

        Raises
        ------
        TypeError
            If `guild_id` or `channel_id` aren't `hikari.Snowflakeish`.
        ValueError
            If neither of `guild_id` or `channel_id` are provided.
        """

        if not guild_id and not channel_id:
            error: str = "At least guild_id or channel_id must be defined"
            raise ValueError(error)
        
        if guild_id and not isinstance(guild_id, hikari.Snowflakeish):
            error: str = "Provided guild ID must be of type `hikari.Snowflakeish`"
            raise TypeError(error)
        
        if channel_id and not isinstance(channel_id, hikari.Snowflakeish):
            error: str = "Provided channel ID must be of type `hikari.Snowflakeish`"
            raise TypeError(error)
        
        if channel_id:
            guild_id = self._connectionsr[channel_id]
        
        try:
            return self._connections[guild_id]
        except KeyError:
            return
    
    async def move(
        self,
        channel_id: hikari.Snowflakeish,
        *,
        guild_id: hikari.Snowflakeish | None = None,
        old_channel_id: hikari.Snowflakeish | None = None,
        mute: bool = False,
        deaf: bool = True,
    ) -> VoiceConnection:
        """
        Move to another voice channel.
        
        Parameters
        ----------
        channel_id : hikari.Snowflakeish
            The ID of the channel to move to.
        guild_id : hikari.Snowflakeish | None
            The ID of the guild you're currently in.
        old_channel_id : hikari.Snowflakeish | None
            The ID of the channel you're currently in.
        mute : bool 
            If the bot should be muted upon moving channels.
        deaf : bool
            If the bot should be deafened upon moving channels.
        
        Note
        ----
        Either `guild_id` or `old_channel_id` have to be provided.

        Returns
        -------
        VoiceConnection
            The active connection to the new voice channel, once fully connected.
        
        Raises
        ------
        asyncio.TimeoutError
            If Discord doesn't send a corresponding voice server/state update.
        TypeError
            - If `channel_id`, `old_channel_id`, or `guild_id` aren't `hikari.Snowflakeish`.
            - If `mute` or `deaf` aren't `bool`.
        ValueError
            If neither of `guild_id` or `old_channel_id` are provided.
        """

        if not guild_id and not old_channel_id:
            error: str = "Either `guild_id` or `old_channel_id` have to be provided"
            raise ValueError(error)
        
        if not isinstance(channel_id, hikari.Snowflakeish):
            error: str = "Provided channel ID must be of type `hikari.Snowflakeish`"
            raise TypeError(error)
        
        if old_channel_id and not isinstance(old_channel_id, hikari.Snowflakeish):
            error: str = "Provided old channel ID must be of type `hikari.Snowflakeish`"
            raise TypeError(error)
        
        if guild_id and not isinstance(guild_id, hikari.Snowflakeish):
            error: str = "Provided guild ID must be of type `hikari.Snowflakeish`"
            raise TypeError(error)
        
        if not isinstance(mute, bool):
            error: str = "Provided mute state must be of type `bool`"
            raise TypeError(error)
        
        if not isinstance(deaf, bool):
            error: str = "Provided deaf state must be of type `bool`"
            raise TypeError(error)
        
        if old_channel_id:
            guild_id = self._connectionsr[old_channel_id]
        return await self._connect(guild_id, channel_id, mute, deaf, True)


from __future__ import annotations

from abc import ABC, abstractmethod

import os

__all__ = (
    "AudioSource",
    "BufferAudioSource",
    "FileAudioSource",
    "URLAudioSource",
)

class AudioSource(ABC):
    """Base audio source implementation."""

    @abstractmethod
    def __init__(self) -> None:
        error: str = "AudioSource should only be subclassed"
        raise NotImplementedError(error)

    @abstractmethod
    def __eq__(self, other: object) -> bool:
        error: str = "AudioSource eq cannot be resolved as it should be subclassed"
        raise NotImplementedError(error)
    
    @abstractmethod
    def __hash__(self) -> int:
        error: str = "AudioSource hash cannot be resolved as it should be subclassed"
        raise NotImplementedError(error)

    def __repr__(self) -> str:
        args: list[str] = []
        for key, value in self.__dict__.items():
            args.append(f"{key.lstrip('_')}={value}")

        return f"{self.__class__.__name__}({', '.join(args)})"

class BufferAudioSource(AudioSource):
    """Buffer audio source implementation."""

    __slots__ = ("_buffer", "name", "_volume",)

    def __init__(self, buffer: bytearray | bytes | memoryview, *, name: str | None = None, volume: float | str | None = None) -> None:
        """
        Create a buffered audio source.
        
        Parameters
        ----------
        buffer : bytearray | bytes | memoryview
            The audio data as a buffer.
        name : str | None
            If provided, an internal name used for display purposes.
        volume : float | str | None
            If provided, overrides the player's set/default volume. Can be scaled (`0.5`, `1.0`, `2.0`, etc.) or dB-based (`-3dB`, etc.).
        """

        self._buffer: bytearray | bytes | memoryview = buffer
        self._volume: float | str | None = volume

        self.name: str | None = name
        """The assigned name of this source for display purposes, if provided."""

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, BufferAudioSource): return False
        return self._buffer == other._buffer
    
    def __hash__(self) -> int:
        return hash(self._buffer)
    
    @property
    def buffer(self) -> bytearray | bytes | memoryview:
        """The audio data as a buffer."""
        return self._buffer

    @property
    def volume(self) -> float | str | None:
        """If provided, the overriding volume for this source."""
        return self._volume

class FileAudioSource(AudioSource):
    """File audio source implementation."""

    __slots__ = ("_filepath", "name", "_volume",)

    def __init__(self, filepath: str, *, name: str | None = None, volume: float | str | None = None) -> None:
        """
        Create a file audio source.
        
        Parameters
        ----------
        filepath : str
            The path, absolute or relative, to the audio file.
        name : str | None
            If provided, an internal name used for display purposes.
        volume : float | str | None
            If provided, overrides the player's set/default volume. Can be scaled (`0.5`, `1.0`, `2.0`, etc.) or dB-based (`-3dB`, etc.).
        """

        if not os.path.exists(filepath):
            error: str = f"Provided filepath does not exist: {filepath}"
            raise FileNotFoundError(error)
        
        self._filepath: str = filepath
        self._volume: float | str | None = volume

        self.name: str | None = name
        """The assigned name of this source for display purposes, if provided."""
    
    def __eq__(self, other: object) -> bool:
        if not isinstance(other, FileAudioSource): return False
        return self._filepath == other._filepath

    def __hash__(self) -> int:
        return hash(self._filepath)

    @property
    def filepath(self) -> str:
        """The path, absolute or relative, to the audio file."""
        return self._filepath
    
    @property
    def volume(self) -> float | str | None:
        """If provided, the overriding volume for this source."""
        return self._volume

class URLAudioSource(AudioSource):
    """URL audio source implementation."""

    __slots__ = ("_url", "name", "_volume",)

    def __init__(self, url: str, *, name: str | None = None, volume: float | str | None = None) -> None:
        """
        Create a URL-based audio source.
        
        Parameters
        ----------
        url : str
            The direct URL to an audio source.
        name : str | None
            If provided, an internal name used for display purposes.
        volume : float | str | None
            If provided, overrides the player's set/default volume. Can be scaled (`0.5`, `1.0`, `2.0`, etc.) or dB-based (`-3dB`, etc.).
        """
        
        self._url: str = url
        self._volume: float | str | None = volume

        self.name: str | None = name
        """The assigned name of this source for display purposes, if provided."""
    
    def __eq__(self, other: object) -> bool:
        if not isinstance(other, URLAudioSource): return False
        return self._url == other._url

    def __hash__(self) -> int:
        return hash(self._url)

    @property
    def url(self) -> str:
        """The direct URL to an audio source."""
        return self._url
    
    @property
    def volume(self) -> float | str | None:
        """If provided, the overriding volume for this source."""
        return self._volume


"""
### hikari-wave: `0.1.0a1`\n
A lightweight, native voice implementation for `hikari`-based Discord bots.

**Documentation:** https://hikari-wave.wildevstudios.net/en/0.1.0a1\n
**GitHub:** https://github.com/WilDev-Studios/hikari-wave
"""

from hikariwave.audio.player import AudioPlayer
from hikariwave.audio.source import *
from hikariwave.client import VoiceClient
from hikariwave.connection import VoiceConnection
from hikariwave.event.events import *
from hikariwave.internal.error import *
from hikariwave.internal.result import Result, ResultReason


from __future__ import annotations

from hikariwave.internal.constants import Audio
from hikariwave.audio.source import (
    AudioSource,
    BufferAudioSource,
    FileAudioSource,
    URLAudioSource,
)
from typing import TypeAlias, TYPE_CHECKING

import asyncio
import logging
import os
import time

if TYPE_CHECKING:
    from hikariwave.client import VoiceClient

Source: TypeAlias = bytearray | bytes | memoryview | str

logger: logging.Logger = logging.getLogger("hikari-wave.ffmpeg")

__all__ = ("FFmpegPool", "FFmpegWorker",)

class FFmpegWorker:
    """Manages a single FFmpeg process when requested."""

    __slots__ = ("_process", "_client",)

    def __init__(self, client: VoiceClient) -> None:
        """
        Create a new worker.

        Parameters
        ----------
        client : VoiceClient
            The voice system client application.
        """

        self._process: asyncio.subprocess.Process = None
        self._client: VoiceClient = client

    async def encode(self, source: AudioSource, output: asyncio.Queue[bytes | None]) -> None:
        """
        Encode an entire audio source and stream each Opus frame into the output.
        
        Parameters
        ----------
        source : AudioSource
            The audio source to read and encode.
        output : asyncio.Queue[bytes | None]
            The output buffer to write each Opus frame into.
        """

        pipeable: bool = False

        if isinstance(source, BufferAudioSource):
            content: bytearray | bytes | memoryview = source._buffer
            pipeable = True
        elif isinstance(source, FileAudioSource):
            content: str = source._filepath
        elif isinstance(source, URLAudioSource):
            content: str = source._url
        else:
            error: str = f"Provided audio source doesn't inherit AudioSource"
            raise TypeError(error)

        volume: float | str = source._volume or 1.0

        args: list[str] = [
            "ffmpeg",
            "-i", "pipe:0" if pipeable else content,
            "-map", "0:a",
            "-af", f"volume={volume}",
            "-acodec", "libopus",
            "-f", "opus",
            "-ar", str(Audio.SAMPLING_RATE),
            "-ac", str(self._client._audio_channels),
            "-b:a", self._client._audio_bitrate,
            "-application", "audio",
            "-frame_duration", str(Audio.FRAME_LENGTH),
            "-loglevel", "warning",
            "pipe:1",
        ]

        self._process = await asyncio.create_subprocess_exec(
            *args,
            stdin=asyncio.subprocess.PIPE if pipeable else None,
            stdout=asyncio.subprocess.PIPE,
            stderr=asyncio.subprocess.DEVNULL,
        )

        if pipeable:
            try:
                self._process.stdin.write(content)
                await self._process.stdin.drain()
                self._process.stdin.close()
                await self._process.stdin.wait_closed()
            except Exception as e:
                logger.error(f"FFmpeg encode error: {e}")
        
        start: float = time.perf_counter()
        while True:
            try:
                header: bytes = await self._process.stdout.readexactly(27)
                if not header.startswith(b"OggS"):
                    return None
                
                segments_count: int = header[26]
                segment_table: bytes = await self._process.stdout.readexactly(segments_count)

                current_packet: bytearray = bytearray()
                for lacing_value in segment_table:
                    data: bytes = await self._process.stdout.readexactly(lacing_value)
                    current_packet.extend(data)

                    if lacing_value < 255:
                        packet_bytes: bytes = bytes(current_packet)

                        if not (
                            packet_bytes.startswith(b"OpusHead") or
                            packet_bytes.startswith(b"OpusTags")
                        ):
                            await output.put(packet_bytes)
                        
                        current_packet.clear()
            except asyncio.IncompleteReadError:
                break
        
        logger.debug(f"FFmpeg finished in {(time.perf_counter() - start) * 1000:.2f}ms")

        await output.put(None)
        await self.stop()
    
    async def stop(self) -> None:
        """
        Stop the internal process.
        """
        
        if not self._process:
            return
        
        for stream in (self._process.stdin, self._process.stdout, self._process.stderr):
            if stream and hasattr(stream, "_transport"):
                try:
                    stream._transport.close()
                except:
                    pass
        
        if self._process.returncode is None:
            try:
                self._process.kill()
                await self._process.wait()
            except ProcessLookupError:
                pass

        self._process = None

class FFmpegPool:
    """Manages all FFmpeg processes and deploys them when needed."""

    __slots__ = (
        "_client", "_max", "_total", "_min",
        "_available", "_unavailable",
    )

    def __init__(self, client: VoiceClient, max_per_core: int = 2, max_global: int = 16) -> None:
        """
        Create a FFmpeg process pool.
        
        Parameters
        ----------
        client : VoiceClient
            The voice system client application.
        max_per_core : int
            The maximum amount of processes that can be spawned per logical CPU core.
        max_global : int
            The maximum, hard-cap amount of processes that can be spawned.
        """
        
        self._client: VoiceClient = client
        self._enabled: bool = True

        self._max: int = min(max_global, os.cpu_count() * max_per_core)
        self._total: int = 0
        self._min: int = 0

        self._available: asyncio.Queue[FFmpegWorker] = asyncio.Queue()
        self._unavailable: set[FFmpegWorker] = set()
    
    async def submit(self, source: AudioSource, output: asyncio.Queue[bytes | None]) -> None:
        """
        Submit and schedule an audio source to be encoded into Opus and stream output into a buffer.
        
        Parameters
        ----------
        source : AudioSource
            The audio source to read and encode.
        output : asyncio.Queue[bytes | None]
            The buffer to stream Opus frames into.
        """
        
        if not self._enabled: return

        if self._available.empty() and self._total < self._max:
            worker: FFmpegWorker = FFmpegWorker(self._client)
            self._total += 1
        else:
            worker: FFmpegWorker = await self._available.get()

        self._unavailable.add(worker)

        async def _run() -> None:
            try:
                await worker.encode(source, output)
            finally:
                self._unavailable.remove(worker)

                if self._total > self._min:
                    self._total -= 1
                else:
                    await self._available.put(worker)

        asyncio.create_task(_run())
    
    async def stop(self) -> None:
        """
        Stop future scheduling and terminate every worker process.
        """

        self._enabled = False

        await asyncio.gather(
            *(unavailable.stop() for unavailable in self._unavailable)
        )
        self._available = asyncio.Queue()
        self._unavailable.clear()

        self._total = 0